
# (c) COPYRIGHT URI/MIT 1997
# Please read the full copyright statement in the file COPYRIGHT.
#
# Authors:
#      jhrg,jimg       James Gallagher (jgallagher@gso.uri.edu)
#
# Clean the cache directory. Removed oldest files until $cache contains less
# than cache_max_size megabytes.
#
# Example use:
# $cache_size = purge_cache(".", 30);
# print "cache now occupies $cache_size bytes\n";

package DODS_Cache;
require Exporter;
@ISA = qw(Exporter);
@EXPORT = qw(purge_cache is_compressed decompress_and_cache);

my $test = 0;
my $debug = 0;

# This regex is used to recognize files that are compressed.
my $compressed_regex = "(\.gz|\.Z)";

use strict;

# Manage the cache directory. Given the pathname to the cache directory and
# the max size (in GB) of the cache, purge files until the size of cached
# files falls below the max size. Use an `oldest first' purge strategy. The
# cache directory can be used by other programs; this routine removes only
# files added by the routine decompress_and_cache(). It does this by
# selecting files which containt he regex `dods_cache#'.
#
# The arguments to this subroutine are: 
# CACHE: the directory that holds the cached files
# CACHE_MAX_SIZE: prune files from the cache until the size (in megabytes) of
# the cache is less than this value.

sub purge_cache {
    my($cache, $cache_max_size) = @_;

    $cache_max_size *= 1048576;	# Bytes/Meg
    my $cached_names_regex = ".*dods_cache#.*";
    my $cache_size = 0;
    
    # Read the cache directory contents.
    opendir CACHE, $cache or die "Could not open the cache directory!\n";
    my(@files) = grep(/$cached_names_regex/, readdir(CACHE));
    closedir CACHE;

    my(%links, %sizes, %times, $nlink, $size, $atime, $file, $d);
    foreach $file (@files) {
        $file = "$cache\/$file";

	($d, $d, $d, $nlink, $d, $d, $d, $size, $atime) = stat($file);
	$links{$file} = $nlink;
	$sizes{$file} = $size;
	$times{$file} = $atime;
	$cache_size += $size;
    }

    print STDERR "Spool size: $cache_size\n" if $debug;
    print STDERR "Spool max size: $cache_max_size\n" if $debug;

    # Remove oldest files first. Continue removing files until cache size
    # falls below max size. Note that for sort ``... the normal calling code
    # for subroutines is bypassed... .'' Look in Wall's ``Programming PERL'',
    # p.218. 7/10/2001 jhrg
    my(@files_sorted_by_times) 
	= sort({$times{$a} <=> $times{$b}} keys(%times));
    
    print STDERR "files sorted by times: @files_sorted_by_times\n" if $debug;

    my $sorted_file;
    foreach $sorted_file (@files_sorted_by_times) {
	# If enough files have been removed, stop flushing the cache.
	if ($cache_size < $cache_max_size) {
	    return $cache_size;
	}
	# Remove the file if only one link to it exists (i.e., not in use).
	# Then update cache_size.
	if ($links{$sorted_file} == 1) {
	    unlink $sorted_file;
	    # print STDERR "Removing $sorted_file\n";
	    $cache_size -= $sizes{$sorted_file};
	}
    }
}

# Return 1 if the filename looks like one we think is compressed, 0
# otherwise. Compressed files end with either .Z or .gz.
sub is_compressed {
    my $file_name = shift;

    return ($file_name =~ m/.*$compressed_regex/) ? 1 : 0;
}

# Given the full path name to a compressed file, decompress that file and
# store the result in the cache. If the file already exists in the cache, do
# nothing. Return the name of the cache entity. Use this in place of the
# pathname to the compressed file. Before calling this routine, make sure
# $pathname has been sanitized!
sub decompress_and_cache {
    my $pathname = shift;
    my $cache_dir = shift;

    # Strip shell metacharacters from the $pathname. 7/10/2001 jhrg
    print STDERR "pathname: $pathname\n" if $debug;
    $pathname !~ m@.*[\s%&()*?<>]+.*@ 
	or return ("", "Found shell meta characters in pathname");

    my $cache_entity = cache_name($pathname, $cache_dir);

    # Only uncompress and cache if the data file actually exits *and* has not
    # already been decompressed and cached.
    if ((! -e $cache_entity) && (-e $pathname)) { 
	# my $uncomp = "gzip -c -d " . $pathname . " > " .  $cache_entity;
	# Using system opens a potential hole. 5/31/2001 jhrg
	# system($uncomp); # uncompress and put output to cache
	
	# This code uses two opens and is a safer than using system since
	# $cache_entity is not run through the shell. However, it sounds like
	# (see ``Programming PERL'') that $pathanme is still exposed to the
	# shell since open uses /bin/sh when a pipe symbol is present. 
	my $uncomp = "gzip -c -d " . $pathname . " |";
	my $buf;
	open GZIP, $uncomp 
	    or return ("", "Could not decompress $pathname: Unable to open the decompressor");
	open DEST, ">$cache_entity" 
	    or return ("", "Could not decompress $pathname: Unable to open destination");
	while (read GZIP, $buf, 16384) {
	    print DEST $buf;
	}
	close GZIP;
	close DEST;
    }

    return ($cache_entity, "");
}

# Private. Build up the cache filename.
sub cache_name {
    my $pathname = shift;
    my $cache_dir = shift;

    $pathname =~ s@^/@@;  # delete leading /
    $pathname =~ s@/@\#@g; # turn remaining / into #
    $pathname =~ s@$compressed_regex@@; # delete trailing filetype

    return $cache_dir . "/" . "dods_cache#" . $pathname;
}

if ($test) {
    (1 == is_compressed("myfile.Z")) || die;
    (1 == is_compressed("my.file.Z")) || die;
    (1 == is_compressed("test.gz")) || die;
    (0 == is_compressed("myfile.")) || die;
    (0 == is_compressed("my.file")) || die;
    (0 == is_compressed("test")) || die;
    (0 == is_compressed("gz")) || die;
    print "\t is_compressed passed all tests\n";

    # NB: Never call cache_name with "" for the cache directory. I'm doing it
    # here just to test stuff. 10/18/2000 jhrg
    print "cache_name myfile: ", cache_name("myfile", ""), "\n" 
  	if $debug; 
    ("/dods_cache#myfile" eq cache_name("myfile", "")) || die;
    ("/dods_cache#myfile" eq cache_name("myfile.Z", "")) || die;
    ("/dods_cache#myfile" eq cache_name("myfile.gz", "")) || die;
    print "cache_name ./myfile.gz", cache_name("./myfile.gz", ""), "\n" 
  	if $debug; 
    ("/dods_cache#this#is#myfile" eq cache_name("/this/is/myfile.Z", "")) || die;
    ("/usr/tmp/dods_cache#myfile" eq cache_name("myfile", "/usr/tmp")) || die;
    ("/usr/tmp/dods_cache#myfile" eq cache_name("myfile.Z", "/usr/tmp")) || die;
    ("/usr/tmp/dods_cache#myfile" eq cache_name("myfile.gz", "/usr/tmp")) || die;
    ("/usr/tmp/dods_cache#this#is#myfile" eq cache_name("/this/is/myfile.Z",
					     "/usr/tmp")) || die;
    ("/usr/tmp/dods_cache#this#is#myfile.HDF" eq cache_name("/this/is/myfile.HDF.Z",
					     "/usr/tmp")) || die;
    ("/usr/tmp/dods_cache#this#is#my.file.HDF" eq cache_name("/this/is/my.file.HDF.Z",
					     "/usr/tmp")) || die;
    print "\t cache_name passed all tests\n";

    ($a, $b) = decompress_and_cache("test_file.gz", "/usr/tmp");
    ("/usr/tmp/dods_cache#test_file" eq $a && "" eq $b) || die;
    (-e "/usr/tmp/dods_cache#test_file") || die;
    (74 == -s "/usr/tmp/dods_cache#test_file") || die;

    ($a, $b) = decompress_and_cache("test/test_file.gz", "/usr/tmp");
    ("/usr/tmp/dods_cache#test#test_file" eq $a && "" eq $b) || die;
    (-e "/usr/tmp/dods_cache#test#test_file") || die;
    (74 == -s "/usr/tmp/dods_cache#test_file") || die;

    # Test scanning for meta chars before sending pathname to /bin/sh.
    ($a, $b) = decompress_and_cache("test_file.gz%3Bcat%20/etc/passwd%3B", 
				    "/usr/tmp");
    ("" eq $a && "" ne $b) || die; # Any error message passes the test.

    print "\t decompress_and_cache passed all tests\n";

    (148 == purge_cache("/usr/tmp", 1)) || die;
    (0 == purge_cache("/usr/tmp", 0)) || die;
    print "\t purge_cache passed all tests\n";

    #unlink "/usr/tmp/dods_cache#test_file"; # get rid of the test files.
    #unlink "/usr/tmp/dods_cache#test#test_file";

    print "All tests succeeded\n";
}

1;

# $Log: DODS_Cache.pm,v $
# Revision 1.4.2.3  2001/10/14 00:42:32  jimg
# Merged with release-3-2-8
#
# Revision 1.4.2.2  2001/07/11 05:08:40  jimg
# Fixed a security problem (Bug 186) where pathnames with shell meta characters
# could make their way to /bin/sh (if they were present in pathnames to be
# decompressed. The shell is still used (haven't figured out how to run gzip in
# the server without redirecting stdout) but paths are sanitized.
# I also fixed a problem with the used of sort. Perl was spitting out a
# warning because the ordering function used by sort used a variable that was
# out of scope.
#
# Revision 1.5  2001/06/15 23:38:36  jimg
# Merged with release-3-2-4.
#
# Revision 1.4.2.1  2001/06/15 17:20:03  jimg
# Added comment about a potential security hole.
#
# Revision 1.4  2000/10/19 17:17:10  jimg
# Added functions to complete the caching system. It is still pretty basic...
#
# Revision 1.3  1999/05/28 05:37:11  jimg
# Merge with alpha-3-0-0
#
# Revision 1.2.26.1  1999/05/27 19:59:53  rmorris
# Refer to cache files using the full path instead a relative path so that
# failure to "be in" the right directory during purging of cache files
# doesn't make cache purging fail and bring down the server when the
# filesystem fills.   This only effects the hdf server.
#
# Revision 1.2  1997/10/09 22:19:17  jimg
# Resolved conflicts in merge of 2.14c to trunk.
#
# Revision 1.1  1997/09/30 22:16:50  jimg
# Created. Initially for use with the JPL Pathfinder server.
